package condor.gahp.gass;

import condor.gahp.*;

import java.io.File;
import java.io.FileInputStream;
import java.io.ByteArrayOutputStream;


import org.ietf.jgss.GSSCredential;
import org.ietf.jgss.GSSException;


import org.globus.io.gass.server.GassServer;


public class GassServerInitHandler implements CommandHandler {

    private GahpInterface gahp;

    public void setGahp (GahpInterface gahp) {
        this.gahp = gahp;
    }

    public CommandHandlerResponse handleCommand (String[] cmd) {

        String fileName = null;
        Integer requestId = null;
		Integer port = null;

        try {
            requestId = new Integer(cmd[1]);
			if (cmd.length > 2) {
				port = new Integer(cmd[2]);
			}
        }
        catch (Exception e) {
            e.printStackTrace(System.err);
            return CommandHandlerResponse.SYNTAX_ERROR;
        }

        return new CommandHandlerResponse (CommandHandlerResponse.SUCCESS,
                      new GassServerInitRunnable (requestId.intValue(),
												  port,
												  gahp));
    } // handleCommand

    class GassServerInitRunnable implements Runnable {
		public Integer port;
        public final int requestId;
        private GahpInterface gahp;
        private GassServer gassServer = null;
        
        public GassServerInitRunnable(int requestId,
									  Integer port,
                                      GahpInterface gahp) {
            this.requestId = requestId;
			this.port = port;
            this.gahp = gahp;
        }
            
        public void run() {
            try {
					// TODO - should we set the credential?

				if (this.port != null && port.intValue() > 0) {
					gassServer = new GassServer(port.intValue());
				} else {
					gassServer = new GassServer();
				}
                gassServer.setOptions (
                                       GassServer.READ_ENABLE |
                                       GassServer.WRITE_ENABLE);
                gassServer.registerDefaultDeactivator();
                gahp.addCleanupStep (
                                     new CleanupStep () {
                                         public void doCleanup() {
                                             if (gassServer != null)
                                                 gassServer.shutdown();
                                         }
                                     });
            }
            catch (Exception e) {
                e.printStackTrace(System.err);
                gahp.addResult (
                                requestId,
                                new String[] { "1", "null"}
                                ); 
            }
                
            gahp.addResult (requestId, new String[] { "0", gassServer.getURL()});
        }
    }
}
