/*
This file is licensed under the terms of the Globus Toolkit Public
License, found at http://www.globus.org/toolkit/download/license.html.
*/

package condor.gahp.gt4;

import condor.gahp.*;


import javax.xml.rpc.ServiceException;
import javax.xml.rpc.Stub;
import javax.xml.namespace.QName;

import org.apache.axis.message.MessageElement;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.axis.gsi.GSIConstants;

import org.apache.axis.message.addressing.EndpointReferenceType;

//import org.globus.exec.client.utils.ManagedJobFactoryHelper;
//import org.globus.exec.client.utils.ManagedJobHelper;
import org.globus.exec.generated.CreateManagedJobInputType;
import org.globus.exec.generated.CreateManagedJobOutputType;
import org.globus.exec.generated.FaultType;
import org.globus.exec.generated.FilePairType;
import org.globus.exec.generated.JobDescriptionType;
import org.globus.exec.generated.ManagedJobFactoryPortType;
import org.globus.exec.generated.ManagedJobPortType;
import org.globus.exec.generated.StateChangeNotificationMessageType;
import org.globus.exec.generated.StateEnumeration;
import org.globus.exec.generated.StateEnumeration;
//import org.globus.exec.service.ManagedJobConstants;
//import org.globus.exec.service.ManagedJobFactoryConstants;
import org.globus.exec.utils.rsl.RSLHelper;
import org.globus.exec.utils.rsl.RSLParseException;
import org.globus.exec.utils.FaultUtils;

import org.globus.security.gridmap.GridMap;

import org.globus.wsrf.WSRFConstants;
import org.globus.wsrf.NotifyCallback;
import org.globus.wsrf.NotificationConsumerManager;
import org.globus.wsrf.NoSuchResourceException;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.authorization.Authorization;
import org.globus.wsrf.impl.security.authorization.HostAuthorization;
import org.globus.wsrf.impl.security.authorization.IdentityAuthorization;
import org.globus.wsrf.impl.security.authorization.SelfAuthorization;
import org.globus.wsrf.impl.security.descriptor.GSISecureMsgAuthMethod;
import org.globus.wsrf.impl.security.descriptor.ResourceSecurityDescriptor;
import org.globus.wsrf.utils.XmlUtils;
import org.globus.exec.generated.FaultResourcePropertyType;

import org.ietf.jgss.GSSCredential;
import org.ietf.jgss.GSSException;

import org.oasis.wsrf.faults.BaseFaultType;

import org.w3c.dom.Element;
import org.w3c.dom.Text;

import java.util.List;
import java.util.Vector;
import java.util.Iterator;

import org.globus.gram.internal.GRAMConstants;

public class JobListener implements NotifyCallback {


    public EndpointReferenceType notificationProducerEPR;
	public EndpointReferenceType notificationConsumerEPR;
    public final GahpInterface gahp;
    public final int requestId;
    public String jobContact = null;
    public String callbackId;
	public boolean queueMessages = true;

	private Vector temporaryMessageQueue = new Vector();

    public JobListener(int requestId, String jobContact, GahpInterface gahp) {
        this.gahp=gahp;
        this.requestId = requestId;
        this.jobContact = jobContact;

		queueMessages = (jobContact == null);
    }

	public void setNotificationConsumerEPR (EndpointReferenceType epr) {
		this.notificationConsumerEPR = epr;
	}

	public EndpointReferenceType getNotificationConsumerEPR () {
		return 	this.notificationConsumerEPR;
	}

	public void setNotificationProducerEPR (EndpointReferenceType epr) {
		this.notificationProducerEPR = epr;
	}

	public EndpointReferenceType getNotificationProducerEPR () {
		return 	this.notificationProducerEPR;
	}

	public String getJobContact() {
		return jobContact;
	}

	public void setJobContact (String jobContact) {
		this.jobContact = jobContact;

		synchronized (temporaryMessageQueue) {	
			queueMessages = (jobContact == null);

			if ((!queueMessages) && (temporaryMessageQueue.size() > 0)) {
					// Flush queue
				Iterator iter=temporaryMessageQueue.iterator();
				while (iter.hasNext()) {
					StateFaultTuple message = 
						(StateFaultTuple)iter.next();
					reportMessage (message);
				} // elihw
				
				temporaryMessageQueue.clear();
			}
		} // synchronized
	}

	private void enqueueMessage(StateFaultTuple message) {
		synchronized (temporaryMessageQueue) {
			if (queueMessages) {
				temporaryMessageQueue.add (message);
			} else {
				reportMessage (message);
			}
		}
	}

	private void reportMessage (StateFaultTuple message) {
		gahp.addResult (
						requestId,
						new String[] {
							jobContact,
							message.state,
							message.faultString,
                            message.exitCode
                           }
						);
	}


	public void deliver (List topicPath,
						 EndpointReferenceType producer,
						 Object message) {
		
		String jobState = null;
        int jobExitCode = -1;
		BaseFaultType jobFault = null;

		try {
			StateChangeNotificationMessageType changeNotification =
                (StateChangeNotificationMessageType)ObjectDeserializer.toObject(
                    (Element) message,
                    StateChangeNotificationMessageType.class);
            StateEnumeration state = changeNotification.getState();
            jobState = state.getValue();
			if (state.equals (StateEnumeration.Failed)) {
				jobFault = getFaultFromRP(changeNotification.getFault());
			}
            jobExitCode = changeNotification.getExitCode();
        }
        catch (Exception e) {
            // TODO Print out the exception
            System.err.println( "Exception while processing callback" );
            return;
		}

        if ( jobFault != null ) {
            System.err.println( "Full fault for job " + jobContact + ":" );
            System.err.println( FaultUtils.faultToString( jobFault ) );
        }

		enqueueMessage (
			new StateFaultTuple (
				jobState,
				(jobFault!=null)?(jobFault.getDescription()[0]).toString():"NULL",
                (jobExitCode==-1) ? "NULL" : ""+jobExitCode));
        
    }

		// This is copied from GramJob.java
    private BaseFaultType getFaultFromRP(FaultResourcePropertyType fault)
    {
        if (fault == null)
        {
            return null;
        }

        if (fault.getFault() != null) {
            return fault.getFault();
        } else if (fault.getCredentialSerializationFault() != null) {
            return fault.getCredentialSerializationFault();
        } else if (fault.getExecutionFailedFault() != null) {
            return fault.getExecutionFailedFault();
        } else if (fault.getFilePermissionsFault() != null) {
            return fault.getFilePermissionsFault();
        } else if (fault.getInsufficientCredentialsFault() != null) {
            return fault.getInsufficientCredentialsFault();
         } else if (fault.getInternalFault() != null) {
            return fault.getInternalFault();
        } else if (fault.getInvalidCredentialsFault() != null) {
            return fault.getInvalidCredentialsFault();
        } else if (fault.getInvalidPathFault() != null) {
            return fault.getInvalidPathFault();
        } else if (fault.getServiceLevelAgreementFault() != null) {
            return fault.getServiceLevelAgreementFault();
        } else if (fault.getStagingFault() != null) {
            return fault.getStagingFault();
        } else if (fault.getUnsupportedFeatureFault() != null) {
            return fault.getUnsupportedFeatureFault();
        } else {
            return null;
        }
    }
    
	public class StateFaultTuple {
		public String state;
		public String faultString;
        public String exitCode;

		public StateFaultTuple(String state, String faultString,
                               String exitCode) {
			this.state=state;
			this.faultString = faultString;
            this.exitCode = exitCode;
		}
	}
}
